@@ -30,6 +30,8 @@
 #include "channels.h"
 #include "queries.h"
 
+#include "irc-nicklist.h"
+#include "irc-queries.h"
 #include "irc-servers-setup.h"
 #include "irc-servers.h"
 #include "channel-rejoin.h"
@@ -55,9 +57,11 @@ void irc_servers_reconnect_deinit(void);
 
 static int cmd_tag;
 
-static int isnickflag_func(char flag)
+static int isnickflag_func(SERVER_REC *server, char flag)
 {
-	return isnickflag(flag);
+	IRC_SERVER_REC *irc_server = (IRC_SERVER_REC *) server;
+
+	return isnickflag(irc_server, flag);
 }
 
 static int ischannel_func(SERVER_REC *server, const char *data)
@@ -98,32 +102,39 @@ static void send_message(SERVER_REC *server, const char *target,
 static void server_init(IRC_SERVER_REC *server)
 {
 	IRC_SERVER_CONNECT_REC *conn;
-	char hostname[100], *address, *ptr, *username;
+	char hostname[100], *address, *ptr, *username, *cmd;
+	GTimeVal now;
 
 	g_return_if_fail(server != NULL);
 
 	conn = server->connrec;
 
 	if (conn->proxy != NULL && conn->proxy_password != NULL &&
-	    *conn->proxy_password != '\0')
-		irc_send_cmdv(server, "PASS %s", conn->proxy_password);
+	    *conn->proxy_password != '\0') {
+		cmd = g_strdup_printf("PASS %s", conn->proxy_password);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
+	}
 
-	if (conn->proxy != NULL && conn->proxy_string != NULL)
-		irc_send_cmdv(server, conn->proxy_string, conn->address, conn->port);
+	if (conn->proxy != NULL && conn->proxy_string != NULL) {
+		cmd = g_strdup_printf(conn->proxy_string, conn->address, conn->port);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
+	}
 
 	if (conn->password != NULL && *conn->password != '\0') {
                 /* send password */
-		server->cmdcount = 0;
-		irc_send_cmdv(server, "PASS %s", conn->password);
+		cmd = g_strdup_printf("PASS %s", conn->password);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
 	}
 
         /* send nick */
-	server->cmdcount = 0;
-	irc_send_cmdv(server, "NICK %s", conn->nick);
+	cmd = g_strdup_printf("NICK %s", conn->nick);
+	irc_send_cmd_now(server, cmd);
+	g_free(cmd);
 
 	/* send user/realname */
-	server->cmdcount = 0;
-
 	if (gethostname(hostname, sizeof(hostname)) != 0 || *hostname == '\0')
 		strcpy(hostname, "xx");
 
@@ -138,26 +149,52 @@ static void server_init(IRC_SERVER_REC *server)
 	}
 
 	/* Replace ':' with '_' in our own hostname (the same IPv6 problem) */
+	/* Replace '\n' and '\r' with '\0' in our own hostname to fix multiple PTR addreses */
 	for (ptr = hostname; *ptr != '\0'; ptr++) {
 		if (*ptr == ':') *ptr = '_';
+		if (*ptr == '\n') *ptr = '\0';
+		if (*ptr == '\r') *ptr = '\0';
 	}
 
+	/* don't allow hostname to begin with number or '+', '-'. those
+	   can be interpreted as ircnet ircd's mode parameter.
+
+	   username/hostname parameters should probably be configurable since
+	   they're only needed with some old servers which uses them to count
+	   unique users. */
+	if ((hostname[0] >= '0' && hostname[0] <= '9') ||
+	    hostname[0] == '+' || hostname[0] == '-')
+		hostname[0] = '_';
+
 	username = g_strdup(conn->username);
 	ptr = strchr(username, ' ');
 	if (ptr != NULL) *ptr = '\0';
 
-	irc_send_cmdv(server, "USER %s %s %s :%s", username, hostname,
-		      address, conn->realname);
+	cmd = g_strdup_printf("USER %s %s %s :%s", username, hostname, address, conn->realname);
+	irc_send_cmd_now(server, cmd);
+	g_free(cmd);
 	g_free(username);
 
-	server->cmdcount = 0;
-
 	if (conn->proxy != NULL && conn->proxy_string_after != NULL) {
-		irc_send_cmdv(server, conn->proxy_string_after,
-			      conn->address, conn->port);
+		cmd = g_strdup_printf(conn->proxy_string_after, conn->address, conn->port);
+		irc_send_cmd_now(server, cmd);
+		g_free(cmd);
 	}
 
+	server->isupport = g_hash_table_new((GHashFunc) g_istr_hash,
+					    (GCompareFunc) g_istr_equal);
+
+	/* set the standards */
+	g_hash_table_insert(server->isupport, g_strdup("CHANMODES"), g_strdup("beIqd,k,lfJ,imnpst"));
+	g_hash_table_insert(server->isupport, g_strdup("PREFIX"), g_strdup("(ohv)@%+"));
+
 	server->cmdcount = 0;
+
+	/* prevent the queue from sending too early, we have a max cut off of 120 secs */
+	/* this will reset to 1 sec after we get the 001 event */
+	g_get_current_time(&now);
+	memcpy(&((IRC_SERVER_REC *)server)->wait_cmd, &now, sizeof(GTimeVal));
+	((IRC_SERVER_REC *)server)->wait_cmd.tv_sec += 120;
 }
 
 SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
@@ -178,7 +215,7 @@ SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
 
 	if (server->connrec->port <= 0) {
 		server->connrec->port =
-			server->connrec->use_ssl ? 9999 : 6667;
+			server->connrec->use_ssl ? 6697 : 6667;
 	}
 
 	server->cmd_queue_speed = ircconn->cmd_queue_speed > 0 ?
@@ -198,6 +235,8 @@ SERVER_REC *irc_server_init_connect(SERVER_CONNECT_REC *conn)
 		ircconn->max_msgs : DEFAULT_MAX_MSGS;
 	server->connrec->use_ssl = conn->use_ssl;
 
+	modes_server_init(server);
+
         server_connect_init((SERVER_REC *) server);
 	return (SERVER_REC *) server;
 }
@@ -268,6 +307,9 @@ static void sig_connected(IRC_SERVER_REC *server)
 	server->isnickflag = isnickflag_func;
 	server->ischannel = ischannel_func;
 	server->send_message = send_message;
+	server->query_find_func =
+		(QUERY_REC *(*)(SERVER_REC *, const char *)) irc_query_find;
+	server->nick_comp_func = irc_nickcmp_ascii;
 
 	server->splits = g_hash_table_new((GHashFunc) g_istr_hash,
 					  (GCompareFunc) g_istr_equal);
@@ -276,6 +318,12 @@ static void sig_connected(IRC_SERVER_REC *server)
 		server_init(server);
 }
 
+static void isupport_destroy_hash(void *key, void *value)
+{
+	g_free(key);
+	g_free(value);
+}
+
 static void sig_disconnected(IRC_SERVER_REC *server)
 {
 	GSList *tmp;
@@ -291,6 +339,12 @@ static void sig_disconnected(IRC_SERVER_REC *server)
 	g_slist_free(server->cmdqueue);
         server->cmdqueue = NULL;
 
+	/* these are dynamically allocated only if isupport was sent */
+	g_hash_table_foreach(server->isupport,
+			     (GHFunc) isupport_destroy_hash, server);
+	g_hash_table_destroy(server->isupport);
+	server->isupport = NULL;
+
 	g_free_and_null(server->wanted_usermode);
 	g_free_and_null(server->real_address);
 	g_free_and_null(server->usermode);
@@ -447,12 +501,12 @@ static int sig_set_user_mode(IRC_SERVER_REC *server)
 	const char *mode;
 	char *newmode, *args;
 
-	if (g_slist_find(servers, server) == NULL)
-		return 0; /* got disconnected */
+	if (!IS_IRC_SERVER(server) || g_slist_find(servers, server) == NULL)
+		return 0; /* not an irc server or got disconnected */
 
 	mode = server->connrec->usermode;
 	newmode = server->usermode == NULL ? NULL :
-		modes_join(server->usermode, mode, FALSE);
+		modes_join(NULL, server->usermode, mode, FALSE);
 
 	if (newmode == NULL || strcmp(newmode, server->usermode) != 0) {
 		/* change the user mode. we used to do some trickery to
@@ -460,7 +514,8 @@ static int sig_set_user_mode(IRC_SERVER_REC *server)
 		   more trouble than worth. (eg. we don't want to remove
 		   some good default server modes, but we don't want to
 		   set back +r, etc..) */
-		args = g_strdup_printf("%s %s", server->nick, mode);
+		args = g_strdup_printf((*mode == '+' || *mode == '-') ? "%s %s" : 
+				       "%s +%s", server->nick, mode);
 		signal_emit("command mode", 3, args, server, NULL);
 		g_free(args);
 	}
@@ -472,6 +527,7 @@ static int sig_set_user_mode(IRC_SERVER_REC *server)
 static void event_connected(IRC_SERVER_REC *server, const char *data, const char *from)
 {
 	char *params, *nick;
+	GTimeVal now;
 
 	g_return_if_fail(server != NULL);
 
@@ -493,6 +549,10 @@ static void event_connected(IRC_SERVER_REC *server, const char *data, const char
 	server->connected = 1;
 	server->real_connect_time = time(NULL);
 
+	/* let the queue send now that we are identified */
+	g_get_current_time(&now);
+	memcpy(&server->wait_cmd, &now, sizeof(GTimeVal));
+
 	if (server->connrec->usermode != NULL) {
 		/* wait a second and then send the user mode */
 		g_timeout_add(1000, (GSourceFunc) sig_set_user_mode, server);
@@ -521,6 +581,109 @@ static void event_server_info(IRC_SERVER_REC *server, const char *data)
 	g_free(params);
 }
 
+static void parse_chanmodes(IRC_SERVER_REC *server, const char *sptr)
+{
+	mode_func_t *modefuncs[] = {
+		modes_type_a,
+		modes_type_b,
+		modes_type_c,
+		modes_type_d
+	};
+	char **item, **chanmodes;
+	int i;
+
+	chanmodes = g_strsplit(sptr, ",", 5); /* ignore extras */
+
+	for (i = 0, item = chanmodes; *item != NULL && i < 4; item++, i++) {
+		unsigned char *p = (unsigned char*) *item;
+		while (*p != '\0') {
+			server->modes[(int)*p].func = modefuncs[i];
+			p++;
+		}
+	}
+
+	g_strfreev(chanmodes);
+}
+
+static void parse_prefix(IRC_SERVER_REC *server, const char *sptr)
+{
+	const char *eptr;
+
+	if (*sptr++ != '(')
+		return; /* Unknown prefix format */
+
+	eptr = strchr(sptr, ')');
+	if (eptr == NULL)
+		return;
+
+	eptr++;
+	while (*sptr != '\0' && *eptr != '\0' && *sptr != ')' && *eptr != ' ') {
+		server->modes[(int)(unsigned char) *sptr].func =
+			modes_type_prefix;
+		server->modes[(int)(unsigned char) *sptr].prefix = *eptr;
+		server->prefix[(int)(unsigned char) *eptr] = *sptr;
+		sptr++; eptr++;
+	}
+}
+
+
+static void event_isupport(IRC_SERVER_REC *server, const char *data)
+{
+	char **item, *sptr, *eptr;
+	char **isupport;
+	gpointer key, value;
+   
+	g_return_if_fail(server != NULL);
+
+	server->isupport_sent = TRUE;
+
+	sptr = strchr(data, ' ');
+	if (sptr == NULL)
+		return;
+	sptr++;
+
+	isupport = g_strsplit(sptr, " ", -1);
+
+	for(item = isupport; *item != NULL; item++) {
+		int removed = FALSE;
+
+		if (**item == '\0')
+			continue;
+
+		if (**item == ':')
+			break;
+
+		sptr = strchr(*item, '=');
+		if (sptr != NULL) {
+			*sptr = '\0';
+			sptr++;
+		}
+
+		eptr = *item;
+		if(*eptr == '-') {
+			removed = TRUE;
+			eptr++;
+		}
+
+		key = value = NULL;
+		if (!g_hash_table_lookup_extended(server->isupport, eptr,
+						  &key, &value) && removed)
+			continue;
+
+		g_hash_table_remove(server->isupport, eptr);
+		if (!removed) {
+			g_hash_table_insert(server->isupport, g_strdup(eptr),
+					    g_strdup(sptr != NULL ? sptr : ""));
+		}
+
+		g_free(key);
+		g_free(value);
+	}
+	g_strfreev(isupport);
+	irc_server_init_isupport(server);
+
+}
+
 static void event_motd(IRC_SERVER_REC *server, const char *data, const char *from)
 {
 	if (server->connected)
@@ -586,6 +749,52 @@ static void event_empty(void)
 {
 }
 
+void irc_server_init_isupport(IRC_SERVER_REC *server)
+{
+	char *sptr;
+	gpointer key, value;
+	/* chanmodes/prefix will fully override defaults */
+	memset(server->modes, 0, sizeof(server->modes));
+	memset(server->prefix, 0, sizeof(server->prefix));
+
+	if ((sptr = g_hash_table_lookup(server->isupport, "CHANMODES")))
+		parse_chanmodes(server, sptr);
+
+	/* This is after chanmode because some servers define modes in both */
+	if (g_hash_table_lookup_extended(server->isupport, "PREFIX",
+					 &key, &value)) {
+		sptr = value;
+		if (*sptr != '(') {
+			/* server incompatible with isupport draft */
+			g_hash_table_remove(server->isupport, key);
+			g_free(key);
+			g_free(value);
+			sptr = NULL;
+		}
+	} else {
+		sptr = NULL;
+	}
+
+	if (sptr == NULL) {
+		sptr = g_strdup("(ohv)@%+");
+		g_hash_table_insert(server->isupport, g_strdup("PREFIX"), sptr);
+	}
+	parse_prefix(server, sptr);
+
+	if ((sptr = g_hash_table_lookup(server->isupport, "MODES"))) {
+		server->max_modes_in_cmd = atoi(sptr);
+		if (server->max_modes_in_cmd < 1)
+			server->max_modes_in_cmd = DEFAULT_MAX_MODES;
+	}
+
+	if ((sptr = g_hash_table_lookup(server->isupport, "CASEMAPPING"))) {
+		if (strstr(sptr, "rfc1459") != NULL)
+			server->nick_comp_func = irc_nickcmp_rfc1459;
+		else
+			server->nick_comp_func = irc_nickcmp_ascii;
+	}
+}
+
 void irc_servers_init(void)
 {
 	settings_add_str("misc", "usermode", DEFAULT_USER_MODE);
@@ -599,6 +808,7 @@ void irc_servers_init(void)
 	signal_add_last("server quit", (SIGNAL_FUNC) sig_server_quit);
 	signal_add("event 001", (SIGNAL_FUNC) event_connected);
 	signal_add("event 004", (SIGNAL_FUNC) event_server_info);
+	signal_add("event 005", (SIGNAL_FUNC) event_isupport);
 	signal_add("event 375", (SIGNAL_FUNC) event_motd);
 	signal_add_last("event 376", (SIGNAL_FUNC) event_end_of_motd);
 	signal_add_last("event 422", (SIGNAL_FUNC) event_end_of_motd); /* no motd */
@@ -623,6 +833,7 @@ void irc_servers_deinit(void)
         signal_remove("server quit", (SIGNAL_FUNC) sig_server_quit);
 	signal_remove("event 001", (SIGNAL_FUNC) event_connected);
 	signal_remove("event 004", (SIGNAL_FUNC) event_server_info);
+	signal_remove("event 005", (SIGNAL_FUNC) event_isupport);
 	signal_remove("event 375", (SIGNAL_FUNC) event_motd);
 	signal_remove("event 376", (SIGNAL_FUNC) event_end_of_motd);
 	signal_remove("event 422", (SIGNAL_FUNC) event_end_of_motd); /* no motd */
