@@ -21,6 +21,7 @@
 #include "module.h"
 #include "signals.h"
 #include "commands.h"
+#include "recode.h"
 #include "network.h"
 #include "net-nonblock.h"
 #include "net-sendbuffer.h"
@@ -104,7 +105,8 @@ CHAT_DCC_REC *dcc_chat_find_id(const char *id)
 	for (tmp = dcc_conns; tmp != NULL; tmp = tmp->next) {
 		CHAT_DCC_REC *dcc = tmp->data;
 
-		if (IS_DCC_CHAT(dcc) && g_strcasecmp(dcc->id, id) == 0)
+		if (IS_DCC_CHAT(dcc) && dcc->id != NULL &&
+		    g_strcasecmp(dcc->id, id) == 0)
 			return dcc;
 	}
 
@@ -146,18 +148,22 @@ void dcc_ctcp_message(IRC_SERVER_REC *server, const char *target,
 		      CHAT_DCC_REC *chat, int notice, const char *msg)
 {
 	char *str;
+	char *recoded;
 
 	if (chat != NULL && chat->sendbuf != NULL) {
 		/* send it via open DCC chat */
+		recoded = recode_out(SERVER(server), msg, chat->nick);
 		str = g_strdup_printf("%s\001%s\001", chat->mirc_ctcp ? "" :
 				      notice ? "CTCP_REPLY " :
-				      "CTCP_MESSAGE ", msg);
+				      "CTCP_MESSAGE ", recoded);
                 dcc_chat_send(chat, str);
 		g_free(str);
 	} else {
+		recoded = recode_out(SERVER(server), msg, target);
 		irc_send_cmdv(server, "%s %s :\001%s\001",
-			      notice ? "NOTICE" : "PRIVMSG", target, msg);
+			      notice ? "NOTICE" : "PRIVMSG", target, recoded);
 	}
+	g_free(recoded);
 }
 
 /* If `item' is a query of a =nick, return DCC chat record of nick */
@@ -195,8 +201,13 @@ static void cmd_msg(const char *data, SERVER_REC *server, WI_ITEM_REC *item)
 	else
 		dcc = NULL;
 
-	if (dcc != NULL && dcc->sendbuf != NULL)
-		dcc_chat_send(dcc, text);
+	if (dcc != NULL && dcc->sendbuf != NULL) {
+		char *recoded;
+
+		recoded = recode_out(server, text, dcc->nick);
+		dcc_chat_send(dcc, recoded);
+		g_free(recoded);
+	}
 
 	if (dcc != NULL || *target == '=')
 		signal_stop();
@@ -204,7 +215,7 @@ static void cmd_msg(const char *data, SERVER_REC *server, WI_ITEM_REC *item)
 	cmd_params_free(free_arg);
 }
 
-static void cmd_me(const char *data, SERVER_REC *server, WI_ITEM_REC *item)
+static void cmd_me(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item)
 {
 	CHAT_DCC_REC *dcc;
 	char *str;
@@ -215,13 +226,13 @@ static void cmd_me(const char *data, SERVER_REC *server, WI_ITEM_REC *item)
 	if (dcc == NULL) return;
 
 	str = g_strconcat("ACTION ", data, NULL);
-	dcc_ctcp_message(NULL, dcc->nick, dcc, FALSE, str);
+	dcc_ctcp_message(server, dcc->nick, dcc, FALSE, str);
 	g_free(str);
 
 	signal_stop();
 }
 
-static void cmd_action(const char *data, SERVER_REC *server)
+static void cmd_action(const char *data, IRC_SERVER_REC *server)
 {
 	CHAT_DCC_REC *dcc;
 	char *target, *text, *str;
@@ -243,7 +254,7 @@ static void cmd_action(const char *data, SERVER_REC *server)
 	dcc = dcc_chat_find_id(target+1);
 	if (dcc != NULL) {
 		str = g_strconcat("ACTION ", text, NULL);
-		dcc_ctcp_message(NULL, dcc->nick, dcc, FALSE, str);
+		dcc_ctcp_message(server, dcc->nick, dcc, FALSE, str);
 		g_free(str);
 	}
 
@@ -251,7 +262,7 @@ static void cmd_action(const char *data, SERVER_REC *server)
 	signal_stop();
 }
 
-static void cmd_ctcp(const char *data, SERVER_REC *server)
+static void cmd_ctcp(const char *data, IRC_SERVER_REC *server)
 {
 	CHAT_DCC_REC *dcc;
 	char *target, *ctcpcmd, *ctcpdata, *str;
@@ -276,7 +287,7 @@ static void cmd_ctcp(const char *data, SERVER_REC *server)
 		g_strup(ctcpcmd);
 
 		str = g_strconcat(ctcpcmd, " ", ctcpdata, NULL);
-		dcc_ctcp_message(NULL, dcc->nick, dcc, FALSE, str);
+		dcc_ctcp_message(server, dcc->nick, dcc, FALSE, str);
 		g_free(str);
 	}
 
@@ -304,8 +315,13 @@ void dcc_chat_input(CHAT_DCC_REC *dcc)
 		}
 
 		if (ret > 0) {
+			char *recoded;
+
 			dcc->transfd += ret;
-			signal_emit("dcc chat message", 2, dcc, str);
+
+			recoded = recode_in(SERVER(dcc->server), str, dcc->nick);
+			signal_emit("dcc chat message", 2, dcc, recoded);
+			g_free(recoded);
 		}
 	} while (ret > 0);
 }
@@ -388,35 +404,77 @@ static void dcc_chat_connect(CHAT_DCC_REC *dcc)
 	}
 }
 
-/* SYNTAX: DCC CHAT [<nick>] */
+static void dcc_chat_passive(CHAT_DCC_REC *dcc)
+{
+	IPADDR own_ip;
+	int port;
+	GIOChannel *handle;
+	char host[MAX_IP_LEN];
+
+	g_return_if_fail(IS_DCC_CHAT(dcc));
+
+	if (dcc->addrstr[0] == '\0' ||
+	    dcc->starttime != 0 || dcc->handle != NULL) {
+		/* already sent a chat request / already chatting */
+		return;
+	}
+
+	handle = dcc_listen(net_sendbuffer_handle(dcc->server->handle),
+			    &own_ip, &port);
+	if (handle == NULL)
+		cmd_return_error(CMDERR_ERRNO);
+
+	dcc->handle = handle;
+	dcc->tagconn = g_input_add(dcc->handle, G_INPUT_READ,
+				   (GInputFunction) dcc_chat_listen, dcc);
+
+	/* Let's send the reply to the other client! */
+	dcc_ip2str(&own_ip, host);
+	irc_send_cmdv(dcc->server, "PRIVMSG %s :\001DCC CHAT CHAT %s %d %d\001",
+		      dcc->nick, host, port, dcc->pasv_id);
+         
+}
+
+/* SYNTAX: DCC CHAT [-passive] [<nick>] */
 static void cmd_dcc_chat(const char *data, IRC_SERVER_REC *server)
 {
 	void *free_arg;
 	CHAT_DCC_REC *dcc;
 	IPADDR own_ip;
-        GIOChannel *handle;
+	GIOChannel *handle;
+	GHashTable *optlist;
+	int p_id;
 	char *nick, host[MAX_IP_LEN];
 	int port;
 
 	g_return_if_fail(data != NULL);
 
-	if (!cmd_get_params(data, &free_arg, 1, &nick))
+	if (!cmd_get_params(data, &free_arg, 1 | PARAM_FLAG_OPTIONS,
+			    "dcc chat", &optlist, &nick))
 		return;
 
 	if (*nick == '\0') {
 		dcc = DCC_CHAT(dcc_find_request_latest(DCC_CHAT_TYPE));
-		if (dcc != NULL)
-			dcc_chat_connect(dcc);
+		if (dcc != NULL) {
+			if (!dcc_is_passive(dcc))
+				dcc_chat_connect(dcc);
+			else
+				dcc_chat_passive(dcc);
+		}
 		cmd_params_free(free_arg);
 		return;
 	}
 
 	dcc = dcc_chat_find_id(nick);
 	if (dcc != NULL && dcc_is_waiting_user(dcc)) {
-		/* found from dcc chat requests,
-		   we're the connecting side */
-		dcc_chat_connect(dcc);
-		cmd_params_free(free_arg);
+		if (!dcc_is_passive(dcc)) {
+			/* found from dcc chat requests,
+			   we're the connecting side */
+			dcc_chat_connect(dcc);
+		} else {
+			/* We are accepting a passive DCC CHAT. */
+			dcc_chat_passive(dcc);
+		}
 		return;
 	}
 
@@ -424,30 +482,50 @@ static void cmd_dcc_chat(const char *data, IRC_SERVER_REC *server)
 	    dcc->server == server) {
 		/* sending request again even while old request is
 		   still waiting, remove it. */
-                dcc_destroy(DCC(dcc));
+		dcc_destroy(DCC(dcc));
 	}
 
-	/* start listening  */
 	if (!IS_IRC_SERVER(server) || !server->connected)
 		cmd_param_error(CMDERR_NOT_CONNECTED);
 
-	handle = dcc_listen(net_sendbuffer_handle(server->handle),
-			    &own_ip, &port);
-	if (handle == NULL)
-		cmd_param_error(CMDERR_ERRNO);
-
 	dcc = dcc_chat_create(server, NULL, nick, "chat");
-        dcc->handle = handle;
-	dcc->tagconn = g_input_add(dcc->handle, G_INPUT_READ,
-				   (GInputFunction) dcc_chat_listen, dcc);
 
-	/* send the chat request */
-	signal_emit("dcc request send", 1, dcc);
+	if (g_hash_table_lookup(optlist, "passive") == NULL) {
+		/* Standard DCC CHAT... let's listen for incoming connections */
+		handle = dcc_listen(net_sendbuffer_handle(server->handle),
+				    &own_ip, &port);
+		if (handle == NULL)
+			cmd_param_error(CMDERR_ERRNO);
 
-	dcc_ip2str(&own_ip, host);
-	irc_send_cmdv(server, "PRIVMSG %s :\001DCC CHAT CHAT %s %d\001",
-		      nick, host, port);
+		dcc->handle = handle;
+		dcc->tagconn =
+			g_input_add(dcc->handle, G_INPUT_READ,
+				    (GInputFunction) dcc_chat_listen, dcc);
 
+		/* send the chat request */
+		signal_emit("dcc request send", 1, dcc);
+
+		dcc_ip2str(&own_ip, host);
+		irc_send_cmdv(server, "PRIVMSG %s :\001DCC CHAT CHAT %s %d\001",
+			      nick, host, port);
+	} else {
+		/* Passive protocol... we want the other side to listen */
+		/* send the chat request */
+		dcc->port = 0;
+		signal_emit("dcc request send", 1, dcc);
+
+		/* generate a random id */
+		srand(time(NULL));
+		p_id = rand() % 64;
+		dcc->pasv_id = p_id;
+
+		/* 16843009 is the long format of 1.1.1.1, we use a fake IP
+		   since the other side shouldn't care of it: they will send
+		   the address for us to connect to in the reply */
+		irc_send_cmdv(server,
+			      "PRIVMSG %s :\001DCC CHAT CHAT 16843009 0 %d\001",
+			      nick, p_id);
+	}
 	cmd_params_free(free_arg);
 }
 
@@ -541,16 +619,18 @@ static void ctcp_msg_dcc_chat(IRC_SERVER_REC *server, const char *data,
         CHAT_DCC_REC *dcc;
 	char **params;
 	int paramcount;
-        int autoallow = FALSE;
-
+        int passive, autoallow = FALSE;
+	
         /* CHAT <unused> <address> <port> */
+	/* CHAT <unused> <address> 0 <id> (DCC CHAT passive protocol) */
 	params = g_strsplit(data, " ", -1);
 	paramcount = strarray_length(params);
 
 	if (paramcount < 3) {
 		g_strfreev(params);
-                return;
+		return;
 	}
+	passive = paramcount == 4 && strcmp(params[2], "0") == 0;
 
 	dcc = DCC_CHAT(dcc_find_request(DCC_CHAT_TYPE, nick, NULL));
 	if (dcc != NULL) {
@@ -559,24 +639,50 @@ static void ctcp_msg_dcc_chat(IRC_SERVER_REC *server, const char *data,
 			   dcc chat from us .. allow it. */
 			dcc_destroy(DCC(dcc));
 			autoallow = TRUE;
-		} else {
+		} else if (!dcc_is_passive(dcc)) {
 			/* we already have one dcc chat request
 			   from this nick, remove it. */
-                        dcc_destroy(DCC(dcc));
+			dcc_destroy(DCC(dcc));
+		} else if (passive) {
+			if (dcc->pasv_id != atoi(params[3])) {
+				/* IDs don't match! */
+				dcc_destroy(DCC(dcc));
+			} else {
+				/* IDs are ok! Update address and port and
+				   connect! */
+				dcc->target = g_strdup(target);
+				dcc->port = atoi(params[2]);
+				dcc_str2ip(params[1], &dcc->addr);
+				net_ip2host(&dcc->addr, dcc->addrstr);
+
+				dcc_chat_connect(dcc);
+				g_strfreev(params);
+				return;
+			}
 		}
 	}
-
+	
 	dcc = dcc_chat_create(server, chat, nick, params[0]);
 	dcc->target = g_strdup(target);
 	dcc->port = atoi(params[2]);
+	
+	if (passive)
+		dcc->pasv_id = atoi(params[3]);
+	
 	dcc_str2ip(params[1], &dcc->addr);
 	net_ip2host(&dcc->addr, dcc->addrstr);
 
 	signal_emit("dcc request", 2, dcc, addr);
 
-	if (autoallow || DCC_CHAT_AUTOACCEPT(dcc, server, nick, addr))
-		dcc_chat_connect(dcc);
-
+	if (autoallow || DCC_CHAT_AUTOACCEPT(dcc, server, nick, addr)) {
+		if (passive) {
+			/* Passive DCC... let's set up a listening socket
+			   and send reply back */
+			dcc_chat_passive(dcc);
+		} else {
+			dcc_chat_connect(dcc);
+		}
+	}
 	g_strfreev(params);
 }
 
@@ -674,6 +780,7 @@ static void event_nick(IRC_SERVER_REC *server, const char *data,
 	char *params, *nick, *tag;
 
 	g_return_if_fail(data != NULL);
+	g_return_if_fail(orignick != NULL);
 
 	params = event_get_params(data, 1, &nick);
 	if (g_strcasecmp(nick, orignick) == 0) {
@@ -693,6 +800,7 @@ static void event_nick(IRC_SERVER_REC *server, const char *data,
 
                 /* change the id too */
 		g_free(dcc->id);
+		dcc->id = NULL;
 		dcc->id = dcc_chat_get_new_id(nick);
 
 		if (query != NULL) {
@@ -716,6 +824,7 @@ void dcc_chat_init(void)
 	command_bind("action", NULL, (SIGNAL_FUNC) cmd_action);
 	command_bind("ctcp", NULL, (SIGNAL_FUNC) cmd_ctcp);
 	command_bind("dcc chat", NULL, (SIGNAL_FUNC) cmd_dcc_chat);
+	command_set_options("dcc chat", "passive");
 	command_bind("mircdcc", NULL, (SIGNAL_FUNC) cmd_mircdcc);
 	command_bind("dcc close", NULL, (SIGNAL_FUNC) cmd_dcc_close);
 	command_bind("whois", NULL, (SIGNAL_FUNC) cmd_whois);
