@@ -436,6 +436,7 @@ void perl_nick_fill_hash(HV *hv, NICK_REC *nick)
 	hv_store(hv, "op", 2, newSViv(nick->op), 0);
 	hv_store(hv, "halfop", 6, newSViv(nick->halfop), 0);
 	hv_store(hv, "voice", 5, newSViv(nick->voice), 0);
+	hv_store(hv, "other", 5, newSViv(nick->other), 0);
 
 	hv_store(hv, "last_check", 10, newSViv(nick->last_check), 0);
 	hv_store(hv, "send_massjoin", 13, newSViv(nick->send_massjoin), 0);
@@ -565,8 +566,13 @@ static void perl_register_protocol(CHAT_PROTOCOL_REC *rec)
 	chat_type = chat_protocol_lookup(rec->name);
 	g_return_if_fail(chat_type >= 0);
 
+#if GLIB_MAJOR_VERSION < 2
 	name = g_strdup(rec->name);
 	g_strdown(name+1);
+#else
+	name = g_ascii_strdown(rec->name,-1);
+	*name = *(rec->name);
+#endif
 
 	/* window items: channel, query */
 	type = module_get_uniq_id_str("WINDOW ITEM TYPE", "CHANNEL");
